@@ -7,6 +7,20 @@
 
 NTL_START_IMPL
 
+static inline
+void CheckFinite(double *p)
+{
+   if (!IsFinite(p)) Error("LLL_QP: numbers too big...use LLL_XD");
+}
+
+
+static inline
+void CheckFinite(quad_float *p)
+{
+   if (!IsFinite(p)) Error("LLL_QP: numbers too big...use LLL_XD");
+}
+
+
 
 static quad_float InnerProduct(quad_float *a, quad_float *b, long n)
 {
@@ -76,15 +90,15 @@ static void RowTransform(vec_ZZ& A, vec_ZZ& B, const ZZ& MU1)
 #define TR_BND (NTL_FDOUBLE_PRECISION/2.0)
 // Just to be safe!!
 
-static quad_float max_abs(quad_float *v, long n)
+static double max_abs(quad_float *v, long n)
 {
    long i;
-   quad_float res, t;
+   double res, t;
 
    res = 0;
 
    for (i = 1; i <= n; i++) {
-      t = fabs(v[i]);
+      t = fabs(v[i].hi);
       if (t > res) res = t;
    }
 
@@ -117,6 +131,7 @@ static void RowTransformFinish(vec_ZZ& A, quad_float *a, long *in_a)
       }
       else {
          conv(a[i], A(i));
+         CheckFinite(&a[i]);
       }
    }
 }
@@ -129,7 +144,7 @@ static void RowTransformFinish(vec_ZZ& A, quad_float *a, long *in_a)
 
 static void RowTransform(vec_ZZ& A, vec_ZZ& B, const ZZ& MU1, 
                          quad_float *a, quad_float *b, long *in_a, 
-                         quad_float& max_a, quad_float max_b, long& in_float)
+                         double& max_a, double max_b, long& in_float)
 // x = x - y*MU
 {
    static ZZ T, MU;
@@ -141,12 +156,18 @@ static void RowTransform(vec_ZZ& A, vec_ZZ& B, const ZZ& MU1,
    long i;
 
    conv(mu, MU1);
+   CheckFinite(&mu);
 
    if (in_float) {
-      max_a += fabs(mu)*max_b; 
-      if (max_a >= TR_BND) {
+      double mu_abs = fabs(mu);
+      if (mu_abs > 0 && max_b > 0 && (mu_abs >= TR_BND || max_b >= TR_BND)) {
          in_float = 0;
       }
+      else {
+         max_a += mu_abs*max_b;
+         if (max_a >= TR_BND) 
+            in_float = 0;
+      }
    }
 
    if (in_float) {
@@ -489,6 +510,7 @@ long ll_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
    quad_float mu1;
 
    quad_float t1;
+   double dt1;
    ZZ T1;
    quad_float *tp;
 
@@ -531,8 +553,8 @@ long ll_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
    in_vec_mem.SetLength(n+1);
    long *in_vec = in_vec_mem.elts();
 
-   quad_float *max_b;
-   max_b = NTL_NEW_OP quad_float [m+1];
+   double *max_b;
+   max_b = NTL_NEW_OP double [m+1];
    if (!max_b) Error("out of memory in lll_LLL_QP");
 
    for (i = 1; i <= m; i++)
@@ -573,7 +595,7 @@ long ll_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
 
       if (st[k] < st[k+1]) st[k+1] = st[k];
       ComputeGS(B, B1, mu, b, c, k, bound, st[k], buf);
-      if (!IsFinite(&c[k])) Error("LLL_QP: numbers too big...use LLL_XD");
+      CheckFinite(&c[k]);
       st[k] = k;
 
       counter = 0;
@@ -661,14 +683,10 @@ long ll_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
             max_b[k] = max_abs(B1[k], n);
    
             b[k] = InnerProduct(B1[k], B1[k], n);
+            CheckFinite(&b[k]);
 
             ComputeGS(B, B1, mu, b, c, k, bound, 1, buf);
-
-            if (!IsFinite(&b[k]))
-               Error("LLL_QP: numbers too big...use LLL_XD");
-
-            if (!IsFinite(&c[k]))
-               Error("LLL_QP: numbers too big...use LLL_XD");
+            CheckFinite(&c[k]);
 
          }
       } while (Fc1);
@@ -682,7 +700,7 @@ long ll_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
             swap(B(i), B(i+1));
             tp = B1[i]; B1[i] = B1[i+1]; B1[i+1] = tp;
             t1 = b[i]; b[i] = b[i+1]; b[i+1] = t1;
-            t1 = max_b[i]; max_b[i] = max_b[i+1]; max_b[i+1] = t1;
+            dt1 = max_b[i]; max_b[i] = max_b[i+1]; max_b[i+1] = dt1;
             if (U) swap((*U)(i), (*U)(i+1));
          }
 
@@ -714,7 +732,7 @@ long ll_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
                tp = B1[i]; B1[i] = B1[i-1]; B1[i-1] = tp;
                tp = mu[i]; mu[i] = mu[i-1]; mu[i-1] = tp;
                t1 = b[i]; b[i] = b[i-1]; b[i-1] = t1;
-               t1 = max_b[i]; max_b[i] = max_b[i-1]; max_b[i-1] = t1;
+               dt1 = max_b[i]; max_b[i] = max_b[i-1]; max_b[i-1] = dt1;
                if (U) swap((*U)(i), (*U)(i-1));
             }
    
@@ -732,7 +750,7 @@ long ll_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
          tp = B1[k]; B1[k] = B1[k-1]; B1[k-1] = tp;
          tp = mu[k]; mu[k] = mu[k-1]; mu[k-1] = tp;
          t1 = b[k]; b[k] = b[k-1]; b[k-1] = t1;
-         t1 = max_b[k]; max_b[k] = max_b[k-1]; max_b[k-1] = t1;
+         dt1 = max_b[k]; max_b[k] = max_b[k-1]; max_b[k-1] = dt1;
          if (U) swap((*U)(k), (*U)(k-1));
 
          k--;
@@ -810,15 +828,16 @@ long LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
 
 
    for (i = 1; i <=m; i++)
-      for (j = 1; j <= n; j++) 
+      for (j = 1; j <= n; j++) {
          conv(B1[i][j], B(i, j));
+         CheckFinite(&B1[i][j]);
+      }
 
 
          
    for (i = 1; i <= m; i++) {
       b[i] = InnerProduct(B1[i], B1[i], n);
-      if (!IsFinite(&b[i]))
-         Error("LLL_QP: numbers too big...use LLL_XD");
+      CheckFinite(&b[i]);
    }
 
 
@@ -970,8 +989,9 @@ void ComputeBKZThresh(quad_float *c, long beta)
 
 
 static 
-void BKZStatus(double tt, double enum_time, long NumIterations, 
-               long NumTrivial, long NumNonTrivial, long NumNoOps, long m, 
+void BKZStatus(double tt, double enum_time, unsigned long NumIterations, 
+               unsigned long NumTrivial, unsigned long NumNonTrivial, 
+               unsigned long NumNoOps, long m, 
                const mat_ZZ& B)
 {
    cerr << "---- BKZ_QP status ----\n";
@@ -1130,14 +1150,15 @@ long BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
 
 
    for (i = 1; i <=m; i++)
-      for (j = 1; j <= n; j++) 
+      for (j = 1; j <= n; j++) {
          conv(B1[i][j], B(i, j));
+         CheckFinite(&B1[i][j]);
+      }
 
          
    for (i = 1; i <= m; i++) {
       b[i] = InnerProduct(B1[i], B1[i], n);
-      if (!IsFinite(&b[i]))
-         Error("BKZ_FD: numbers too big...use BKZ_XD");
+      CheckFinite(&b[i]);
    }
 
    // cerr << "\n";
@@ -1149,10 +1170,10 @@ long BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
    double tt;
 
    double enum_time = 0;
-   long NumIterations = 0;
-   long NumTrivial = 0;
-   long NumNonTrivial = 0;
-   long NumNoOps = 0;
+   unsigned long NumIterations = 0;
+   unsigned long NumTrivial = 0;
+   unsigned long NumNonTrivial = 0;
+   unsigned long NumNoOps = 0;
 
    long verb = verbose;
 
@@ -1369,10 +1390,13 @@ long BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                   t1 = b[i-1]; b[i-1] = b[i]; b[i] = t1;
                }
       
-               for (i = 1; i <= n; i++)
+               for (i = 1; i <= n; i++) {
                   conv(B1[jj][i], B(jj, i));
+                  CheckFinite(&B1[jj][i]);
+               }
       
                b[jj] = InnerProduct(B1[jj], B1[jj], n);
+               CheckFinite(&b[jj]);
       
                if (b[jj] == 0) Error("BKZ_QP: internal error"); 
       
@@ -1636,14 +1660,15 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
 
 
    for (i = 1; i <=m; i++)
-      for (j = 1; j <= n; j++) 
+      for (j = 1; j <= n; j++) {
          conv(B1[i][j], B(i, j));
+         CheckFinite(&B1[i][j]);
+      }
 
          
    for (i = 1; i <= m; i++) {
       b[i] = InnerProduct(B1[i], B1[i], n);
-      if (!IsFinite(&b[i]))
-         Error("BKZ_FD: numbers too big...use BKZ_XD");
+      CheckFinite(&b[i]);
    }
 
    // cerr << "\n";
@@ -1655,10 +1680,10 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
    double tt;
 
    double enum_time = 0;
-   long NumIterations = 0;
-   long NumTrivial = 0;
-   long NumNonTrivial = 0;
-   long NumNoOps = 0;
+   unsigned long NumIterations = 0;
+   unsigned long NumTrivial = 0;
+   unsigned long NumNonTrivial = 0;
+   unsigned long NumNoOps = 0;
 
    long verb = verbose;
 
@@ -1817,7 +1842,7 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
 
             clean = 0;
 
-            // we treat the case that the NTL_NEW_OP vector is b_s (jj < s <= kk)
+            // we treat the case that the new vector is b_s (jj < s <= kk)
             // as a special case that appears to occur most of the time.
    
             s = 0;
@@ -1879,10 +1904,13 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                   t1 = b[i-1]; b[i-1] = b[i]; b[i] = t1;
                }
       
-               for (i = 1; i <= n; i++)
+               for (i = 1; i <= n; i++) {
                   conv(B1[jj][i], B(jj, i));
+                  CheckFinite(&B1[jj][i]);
+               }
       
                b[jj] = InnerProduct(B1[jj], B1[jj], n);
+               CheckFinite(&b[jj]);
       
                if (b[jj] == 0) Error("BKZ_QP: internal error"); 
       
