@@ -37,9 +37,8 @@ void zz_pEX::normalize()
 
    n = rep.length();
    if (n == 0) return;
-   p = rep.elts() + (n-1);
-   while (n > 0 && IsZero(*p)) {
-      p--; 
+   p = rep.elts() + n;
+   while (n > 0 && IsZero(*--p)) {
       n--;
    }
    rep.SetLength(n);
@@ -107,6 +106,7 @@ long operator==(const zz_pEX& a, const zz_pE& b)
 
 
 
+
 void SetCoeff(zz_pEX& x, long i, const zz_pE& a)
 {
    long j, m;
@@ -114,19 +114,25 @@ void SetCoeff(zz_pEX& x, long i, const zz_pE& a)
    if (i < 0) 
       Error("SetCoeff: negative index");
 
-   if (i >= (1L << (NTL_BITS_PER_LONG-4)))
+   if (NTL_OVERFLOW(i, 1, 0))
       Error("overflow in SetCoeff");
 
    m = deg(x);
 
    if (i > m) {
-      long pos = x.rep.position(a);
-      x.rep.SetLength(i+1);
+      /* careful: a may alias a coefficient of x */
 
-      if (pos != -1)
-         x.rep[i] = x.rep.RawGet(pos);
-      else
+      long alloc = x.rep.allocated();
+
+      if (alloc > 0 && i >= alloc) {
+         zz_pE aa = a;
+         x.rep.SetLength(i+1);
+         x.rep[i] = aa;
+      }
+      else {
+         x.rep.SetLength(i+1);
          x.rep[i] = a;
+      }
 
       for (j = m+1; j < i; j++)
          clear(x.rep[j]);
@@ -137,6 +143,7 @@ void SetCoeff(zz_pEX& x, long i, const zz_pE& a)
    x.normalize();
 }
 
+
 void SetCoeff(zz_pEX& x, long i, const zz_p& aa)
 {
    long j, m;
@@ -144,7 +151,7 @@ void SetCoeff(zz_pEX& x, long i, const zz_p& aa)
    if (i < 0)
       Error("SetCoeff: negative index");
 
-   if (i >= (1L << (NTL_BITS_PER_LONG-4)))
+   if (NTL_OVERFLOW(i, 1, 0))
       Error("overflow in SetCoeff");
 
    NTL_zz_pRegister(a);  // watch out for aliases!
@@ -181,7 +188,7 @@ void SetCoeff(zz_pEX& x, long i)
    if (i < 0) 
       Error("coefficient index out of range");
 
-   if (i >= (1L << (NTL_BITS_PER_LONG-4)))
+   if (NTL_OVERFLOW(i, 1, 0))
       Error("overflow in SetCoeff");
 
    m = deg(x);
@@ -579,7 +586,7 @@ void mul(zz_pEX& c, const zz_pEX& a, const zz_pEX& b)
    long n = zz_pE::degree();
    long n2 = 2*n-1;
 
-   if (da+db+1 >= (1L << (NTL_BITS_PER_LONG-4))/n2)
+   if (NTL_OVERFLOW(da+db+1, n2, 0))
       Error("overflow in zz_pEX mul");
 
 
@@ -716,7 +723,7 @@ void sqr(zz_pEX& c, const zz_pEX& a)
    long n = zz_pE::degree();
    long n2 = 2*n-1;
 
-   if (2*da+1 >= (1L << (NTL_BITS_PER_LONG-4))/n2)
+   if (NTL_OVERFLOW(2*da+1, n2, 0))
       Error("overflow in zz_pEX sqr");
 
    long i, j;
@@ -1303,6 +1310,11 @@ void PlainRem(zz_pEX& r, const zz_pEX& a, const zz_pEX& b)
 
 void RightShift(zz_pEX& x, const zz_pEX& a, long n)
 {
+   if (IsZero(a)) {
+      clear(x);
+      return;
+   }
+
    if (n < 0) {
       if (n < -NTL_MAX_LONG) Error("overflow in RightShift");
       LeftShift(x, a, -n);
@@ -1331,20 +1343,22 @@ void RightShift(zz_pEX& x, const zz_pEX& a, long n)
 
 void LeftShift(zz_pEX& x, const zz_pEX& a, long n)
 {
-   if (n < 0) {
-      if (n < -NTL_MAX_LONG) Error("overflow in LeftShift");
-      RightShift(x, a, -n);
-      return;
-   }
-
-   if (n >= (1L << (NTL_BITS_PER_LONG-4)))
-      Error("overflow in LeftShift");
-
    if (IsZero(a)) {
       clear(x);
       return;
    }
 
+   if (n < 0) {
+      if (n < -NTL_MAX_LONG) 
+         clear(x);
+      else
+         RightShift(x, a, -n);
+      return;
+   }
+
+   if (NTL_OVERFLOW(n, 1, 0))
+      Error("overflow in LeftShift");
+
    long m = a.rep.length();
 
    x.rep.SetLength(m+n);
@@ -1383,10 +1397,10 @@ void NewtonInv(zz_pEX& c, const zz_pEX& a, long e)
    zz_pEX g, g0, g1, g2;
 
 
-   g.rep.SetMaxLength(e);
-   g0.rep.SetMaxLength(e);
-   g1.rep.SetMaxLength((3*e+1)/2);
-   g2.rep.SetMaxLength(e);
+   g.rep.SetMaxLength(E[0]);
+   g0.rep.SetMaxLength(E[0]);
+   g1.rep.SetMaxLength((3*E[0]+1)/2);
+   g2.rep.SetMaxLength(E[0]);
 
    conv(g, x);
 
@@ -1422,7 +1436,7 @@ void InvTrunc(zz_pEX& c, const zz_pEX& a, long e)
       return;
    }
 
-   if (e >= (1L << (NTL_BITS_PER_LONG-4)))
+   if (NTL_OVERFLOW(e, 1, 0))
       Error("overflow in InvTrunc");
 
    NewtonInv(c, a, e);
@@ -1441,7 +1455,7 @@ void build(zz_pEXModulus& F, const zz_pEX& f)
 
    if (n <= 0) Error("build(zz_pEXModulus,zz_pEX): deg(f) <= 0");
 
-   if (n >= (1L << (NTL_BITS_PER_LONG-4))/zz_pE::degree())
+   if (NTL_OVERFLOW(n, zz_pE::degree(), 0))
       Error("build(zz_pEXModulus,zz_pEX): overflow");
 
    F.tracevec.SetLength(0);
@@ -2361,7 +2375,7 @@ static
 void ProjectPowers(vec_zz_pE& x, const zz_pEX& a, long k, 
                    const zz_pEXArgument& H, const zz_pEXModulus& F)
 {
-   if (k < 0 || k >= (1L << (NTL_BITS_PER_LONG-4)) || deg(a) >= F.n)
+   if (k < 0 || NTL_OVERFLOW(k, 1, 0) || deg(a) >= F.n)
       Error("ProjectPowers: bad args");
 
    long m = H.H.length()-1;
@@ -2491,8 +2505,8 @@ void BerlekampMassey(zz_pEX& h, const vec_zz_pE& a, long m)
 
 void MinPolySeq(zz_pEX& h, const vec_zz_pE& a, long m)
 {
-   if (m < 0 || m >= (1L << (NTL_BITS_PER_LONG-4))) Error("MinPoly: bad args");
-   if (a.length() < 2*m) Error("BerlekampMassey: sequence too short");
+   if (m < 0 || NTL_OVERFLOW(m, 1, 0)) Error("MinPoly: bad args");
+   if (a.length() < 2*m) Error("MinPoly: sequence too short");
 
    BerlekampMassey(h, a, m);
 }
@@ -2864,8 +2878,8 @@ void PowerXMod(zz_pEX& hh, const ZZ& e, const zz_pEXModulus& F)
 
 void reverse(zz_pEX& x, const zz_pEX& a, long hi)
 {
-   if (hi < -1) Error("reverse: bad args");
-   if (hi >= (1L << (NTL_BITS_PER_LONG-4)))
+   if (hi < 0) { clear(x); return; }
+   if (NTL_OVERFLOW(hi, 1, 0))
       Error("overflow in reverse");
 
    if (&x == &a) {
@@ -3242,7 +3256,7 @@ void ProjectPowersTower(vec_zz_p& x, const vec_zz_pE& a, long k,
 {
    long n = F.n;
 
-   if (a.length() > n || k < 0 || k >= (1L << (NTL_BITS_PER_LONG-4)))
+   if (a.length() > n || k < 0 || NTL_OVERFLOW(k, 1, 0))
       Error("ProjectPowers: bad args");
 
    long m = H.H.length()-1;
